from numpy import *
from numpy.linalg import norm
from matplotlib.pyplot import *
from SplittingParameters import *

def init_pos_vel(d, n, box_width):
    x = linspace(0, box_width, n)
    tx = meshgrid(*[x]*d)
    pos = list(zip(*[xs.flatten() for xs in tx]))
    vel = zeros_like(pos)
    return array(pos), array(vel)


class Solution:
    def __init__(self, x, v):
        self._x = x.copy()
        self._v = v.copy()

    def __getitem__(self, key):
        if key is 'q':
            return self._x
        elif key is 'p':
            return self._v
        else:
            raise KeyError('Invalid key: ' + str(key))

    def __setitem__(self, key, value):
        if key is 'q':
            self._x = value.copy()
        elif key is 'p':
            self._v = value.copy()
        else:
            raise KeyError('Invalid key: ' + str(key))

    def __rmul__(self, val):
        return Solution(self._x*val, self._v*val)

    def __lmul__(self, val):
        return Solution(self._x*val, self._v*val)


def phiT(dt, sol):
    r"""
    dt  -- Zeitschrittweite
    sol -- Aktuelle Loesung
    """
    #########################################################
    #                                                       #
    # TODO: Implementieren Sie den Splitting Operator phiT. #
    #                                                       #
    #########################################################
    return sol

def phiV(dt, sol, F):
    r"""
    dt  -- Zeitschrittweite
    sol -- Aktuelle Loesung
    F   -- Funktion zur Berechnung der Kraft
    """
    #########################################################
    #                                                       #
    # TODO: Implementieren Sie den Splitting Operator phiV. #
    #                                                       #
    #########################################################
    return sol


def small_perturbation(d=2, N=5, box_width=5, T=5, nsteps=100, method='Y4'):
    phi = 2*pi*random.rand(N*N)

    vref =  1e-6 * array([cos(phi), sin(phi)]).T
    x = init_pos_vel(d, N, box_width)[0]

    ## Original Problem
    # inital conditions + solution
    v = vref
    sol = Solution(x, v)
    # initialize splitting
    S = SplittingParameters()

    ###############################################################
    #                                                             #
    # TODO: Verwenden Sie 'S.intsplit(...)' zur Zeitintegration.  #
    #                                                             #
    ###############################################################

    ## Perturbed Problem
    v[0,:] += 1e-10
    solp = Solution(x, v)

    ###############################################################
    #                                                             #
    # TODO: Verwenden Sie 'S.intsplit(...)' zur Zeitintegration.  #
    #                                                             #
    ###############################################################

    # do all the plotting
    compare(ts, sols, solsp)


def compute_energy_times(ts, sols):
    r"""
    ts   -- Sample times
    sols -- Solution at the sample times
    """
    # compute energies
    E = []
    for i in range(len(ts)):
        E.append(compute_energy(sols[i]['q'], sols[i]['p'], LJ))
    return array(E)


def compute_energy(x, v, V):
    r"""
    Keyword Arguments:
    x -- positions
    v -- velocity
    V -- Potential
    """
    Vpot = 0
    Vkin = 0
    for i in range(len(x)):
        Vkin += 0.5*dot(v[i,:],v[i,:])
        for j in range(len(x)):
            if i==j:
                continue
            Vpot += 0.5*V(norm(x[i,:]-x[j,:]))
    return Vpot+Vkin, Vpot, Vkin


def plot_trajectories(ts, sols, sols_perturbed):
    colors = ['r', 'g', 'b', 'm', 'k']*100

    xhist  = array([x['q'] for x in sols])
    vhist  = array([x['p'] for x in sols])
    xhistp = array([x['q'] for x in sols_perturbed])
    vhistp = array([x['p'] for x in sols_perturbed])

    figure()
    for i in range(xhist.shape[1]):
        plot(xhist[0,i,0], xhist[0,i,1],'o', color=colors[i])
        plot(xhist[-1,i,0], xhist[-1,i,1],'s', color=colors[i])
        plot(xhistp[-1,i,0], xhistp[-1,i,1],'d', color=colors[i])
        plot(xhist[:,i,0], xhist[:,i,1], color=colors[i], label='original')
        plot(xhistp[:,i,0], xhistp[:,i,1], '--', color=colors[i], label='perturbed')
        title('Trajektorien')
        xlabel('$x$')
        ylabel('$y$')
        legend( [Line2D([],[],linestyle='--', color='k'),
                 Line2D([],[],linestyle='-', color='k'),
                 Line2D([],[], linestyle='', marker='o', color='k'),
                 Line2D([],[], linestyle='', marker='s', color='k'),
                 Line2D([],[], linestyle='', marker='d', color='k')],
                ["perturbed",
                 "original",
                 "$x(t=0)$",
                 r'$x(t=t_{end})$',
                 r'$x_{\epsilon}(t=t_{end})$'],
                numpoints=1)
    axis('equal')
    grid(True)
    savefig('splitting-trajektorien.png')

    figure()
    di=  [sqrt(sum((xhistp[i]-xhist[i])**2)) for i in range(len(ts))]
    vi = [sqrt(sum((vhistp[i]-vhist[i])**2)) for i in range(len(ts))]
    semilogy(ts, di, label=r'$|| x - x_{\epsilon}||$')
    semilogy(ts, vi, label=r'$|| v- v_{\epsilon}||$')
    xlabel('$t$')
    ylabel('Abweichung')
    grid(True)
    title('Phasenraumdistanz')
    legend(loc='lower right')
    savefig('splitting-phasenraumdistanz.png')


def compare(ts, sols, sols_perturbed):
    E = compute_energy_times(ts, sols)
    Ep = compute_energy_times(ts, sols_perturbed)

    figure()
    suptitle('Original Problem\n\n')
    subplot(121)
    plot(ts, E[:,2], label='T')
    plot(ts, E[:,1], label='U')
    plot(ts, E[:,0], label='E')
    grid(True)
    xlabel('$t$')
    ylabel('Energy')
    legend(loc='best')
    subplot(122)
    title('Conservation of total Energy')
    semilogy(ts, abs(E[0,0]-E[:,0]))
    xlabel('$t$')
    ylabel('$E$')
    grid(True)
    tight_layout()
    savefig('splitting-original.png')

    figure()
    suptitle('Perturbed Problem\n\n')
    subplot(121)
    plot(ts, Ep[:,2], label='T')
    plot(ts, Ep[:,1], label='U')
    plot(ts, Ep[:,0], label='E')
    grid(True)
    xlabel('$t$')
    ylabel('Energy')
    legend(loc='best')
    subplot(122)
    title('Conservation of total Energy')
    semilogy(ts, abs(Ep[0,0]-Ep[:,0]))
    xlabel('$t$')
    ylabel('$E$')
    grid(True)
    tight_layout()
    savefig('splitting-perturbed.png')

    figure()
    semilogy(ts, abs(E[:,0]-Ep[:,0]))
    ylabel(r'$|E-E_{\epsilon}|$')
    xlabel(r'$t$')
    grid(True)
    savefig('splitting-energie-differenz.png')

    plot_trajectories(ts, sols, sols_perturbed)




if __name__ == '__main__':
    epsilon = 1
    sigma = 1
    LJ = lambda r: epsilon* ((sigma/r)**12 - (sigma/r)**6)
    DLJ = lambda x: 6*epsilon*(2*x*(sigma/norm(x))**14 - x*(sigma/norm(x))**8)
    F = lambda x: eval_DV(x, DLJ)

    d = 2
    N = 3
    box_width = 3
    T = 50.0
    nsteps = 1000

    ##################################
    #                                #
    # TODO: Waehlen Sie die Methode. #
    #                                #
    ##################################
    small_perturbation(d, N, box_width, T, nsteps, method='SS')
    #small_perturbation(d, N, box_width, T, nsteps, method='BM42')
